#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Sensor v2.0 MQTT flavour
# (c) Electrosense Project
#
#  [TODO] License text
#
# Héctor Cordobés - hcordobes@imdea.org - 2017


import sys
import time
import datetime
import random
import logging
import subprocess
import errno
import platform
import psutil
import apt_pkg
import paho.mqtt.client as mqtt
import netifaces
import os
import signal
import socket

# internals
import config
import mqttavro as ma
from stypes import ConfigError

# Config init
# Read configuration file
# Beware: log output is not configured yet
cfgs = ["sensor.cfg", "/etc/electrosense/sensor.cfg"]
cp = config.read(cfgs)

# TODO check config is valid
# Config dictionaries
mqttconfig = config.map(cp, "mqtt")
logconfig = config.map(cp, "log")
avroconfig = config.map(cp, "avro")
envconfig = config.map(cp, "environment")
parameterTable = config.map(cp, "parameters")

# ***************** Log
logLevel = getattr(logging, logconfig["level"].upper(), None)
if not isinstance(logLevel, int):
    raise ValueError('Invalid log level: %s' % logconfig["level"])

logger = logging.getLogger("sensor")
logger.setLevel(logLevel)

formatter = logging.Formatter("%(asctime)s:%(name)s:%(levelname)s:%(message)s")

# Console handler
if "console" in logconfig.keys():
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)

# File handler
# XXX Use RotatingFileHandler?
if "file" in logconfig.keys():
    fh = logging.FileHandler(filename=logconfig["file"])
    fh.setFormatter(formatter)
    logger.addHandler(fh)

logger.info("****** MQTT Sensor Starting *******")

subproc = None

def generateConfigList(inputParams):
    """Generate a list of parameters representing the system command to start measurements,
    given a dictionary of parameters as received by the backend.

    Example:
        params = {"command": "run_gpu_sensor", "minFreq": 0, "maxFreq": 500}
        returns ["run_gpu_sensor", "0", "500"]

    The parameters must be present in the [parameters] section of the configuration,
    otherwise they are ignored. Each parameter is substituted with its command line
    argument.
    [2018-09-11] The check for mandatory parameters is left to the process implementation
    and the system call. With this change the logic is simpler for future new processes.
    As the parameter list may not be completely homogeneous a prefix/suffix approach is used.
    """

    # Order is preserved; use lowercase
    prefixParams = ["command"]
    suffixParams = ["minfreq", "maxfreq"]
    inputParams = dict( (k.lower(), v) for k,v in inputParams.items())
    outputParamKeys = []

    # Add prefix params if present
    for param in prefixParams:
        if param in inputParams.keys():
            outputParamKeys.append(param)
    # Add all params which are not prefix or suffix
    for param in inputParams.keys():
        if param not in prefixParams and param not in suffixParams:
            outputParamKeys.append(param)
    # And add suffix params if present
    for param in suffixParams:
        if param in inputParams.keys():
            outputParamKeys.append(param)

    # begin construction of config
    cfgList = []
    if (envconfig["sudo"] == "true"):
        cfgList.append("sudo")

    # from each parameter name, look for its CLI option
    for param in outputParamKeys:
        logger.debug("considering param %s", param)
        # FIXME lower() in ConfigParser should be bypassed
        if param in parameterTable.keys():
            try:
                paramId = parameterTable[param]
                if paramId:
                    cfgList.append(paramId)
                cfgList.append(inputParams[param])
            except:
                logger.error("Boom %s")
        else:
            logger.warning("Unknown option parameter: %s", param)
    logger.debug("Config file contents: %s", str(cfgList))
    return cfgList


def errorFail(cause, newStatus):
    """Create status info with "ERROR" and set global staus to newStatus"""
    global status
    global statusInfo
    global campaign
    status = newStatus
    statusInfo["Status"] = "ERROR"
    statusInfo["Cause"] = cause
    logger.error(cause)
    # TODO Keeping twice the same info???
    campaign = "Unknown"
    statusInfo["CampaignId"] = campaign
    return


def stopRunningProcess():
    """Kill measurement sub process

    Returns:
        True if kill succeeded, False otherwise
    """
    global subproc

    logger.debug("Stopping running process...")
    if subproc is None:
        logger.debug("No previous subprocess found")
        return

    try:
        # POSIX prevents sending signals to other user process
        # so we cannot kill the parent of a sudo'd process directly from
        # a user-run Python process
        if(envconfig["sudo"] == "true"):
            execName = subproc.args[1]
            logger.debug("Exec name: %s", execName)
            pid = subproc.pid
            subprocess.call(["sudo", "kill", str(pid)])
            logger.debug("Called kill over: %s", str(pid))
            time.sleep(1)
            if subproc.poll() is None:
                logger.debug("Tough one, kill -9")
                subprocess.call(["sudo", "kill", "-9", str(pid)])
        else:
            execName = subproc.args[0]
            subproc.kill()
            time.sleep(1)
            if subproc.poll() is None:
                subproc.terminate()
        # Fallback
        logger.debug("P-Killing process %s", str(execName) )
        subprocess.call(["sudo", "pkill", "-9", execName])
    except OSError as e:
        logger.debug("OSError killing process: %s", str(e.errno))
        # Non-existent process exception is good
        if e.errno != errno.ESRCH:
            errorFail("OS error killing previous sensor"
                      "instance: " + e.strerror, "ERROR")
            sendSensorStatus()
            return False
    subproc = None
    return True


# BIG function following

def processMeasurementCommand(msg):
    """Start or Stop measurement
    Args:
        - msg (dict): containing Command, Uuid, Timestamp and SensingParams

    Command may be either START or STOP
      START: sanitycheck, stop previous, create config file, start new campaign
      STOP: sanity check

    When measurement is started, a message indicating this event is published via MQTT
    (status BUSY).
    """
    global status
    global statusInfo
    global campaign
    global subproc

    # Read message contents
    uuid = msg["Uuid"]
    command = msg["Command"]
    params = msg["SensingParams"]
    timestamp = msg["Timestamp"]
    logger.info("Command: " + command)
    logger.debug("Uuid: " + uuid)
    logger.debug("Params: " + str(params))
    logger.debug("Timestamp: " + str(timestamp))
    # TODO thoughts: should we take Timestamp for anything here? Sync? Sanity?

    # Clean statusInfo
    statusInfo = {"Uuid": uuid}
    if status == "ERROR":
        logger.warning("Trying to " + msg["Command"] +
                       " after an ERROR status")

    if msg["Command"] == "START":
        logger.info("Launching sensor process")

        # stop previous campaign if needed
        if status == "SENSING":
            logger.warning("There was a previous campaign running: %s",
                           str(campaign))

        if subproc is not None:
            # a process was launched previously
            if not stopRunningProcess():
                return

        # Send ACK message as we are to launch
        # a potentially blocking call afterwards
        status = "BUSY"
        sendSensorStatus()
        cfgList = generateConfigList(params)

        # Added feedback info for the executed command
        statusInfo["ExecutedCommand"] = " ".join(cfgList)

        try:
            subproc = subprocess.Popen(cfgList,
                                       cwd="/usr/share/electrosense",
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       universal_newlines=True)
        except OSError as e:
            errorFail("Could not launch process: " + e.strerror, "IDLE")
            sendSensorStatus()
            return

        campaign = params.get("CampaignId", "Unknown")
        status = "SENSING"
        statusInfo["CampaignId"] = campaign
        sendSensorStatus()
        return
    # END of START command

    if msg["Command"] == "STOP":
        if status == "IDLE":
            statusInfo["CampaignId"] = campaign
            sendSensorStatus()
            return
        status = "BUSY"
        sendSensorStatus()
        # stop command
        if stopRunningProcess():
            status = "IDLE"
            campaign = "Unknown"
            sendSensorStatus()
        else:
            return


def processGroupIDs(msg):
    global groups
    logger.debug(msg)
    newgroups = set(msg.get("GroupIDs"))
    unsubscribe = groups - newgroups
    subscribe = newgroups-groups
    for groupid in unsubscribe:
        logger.info("Unsubscribing from group: %s", groupid)
        topic = "controller/sensor/group/"+groupid
        mc.unsubscribe(topic)
    for groupid in subscribe:
        logger.info("Subscribing to group: %s", groupid)
        topic = "controller/sensor/group/"+groupid
        mc.subscribe(topic, 1)
    groups = newgroups


def processStatusRequest(msg):
    """Reply either with SensorInfo or SensorStatus depending on msg["Type"]"""
    if msg["Type"] == "SENSORINFO":
        sendSensorInfo()
        return
    if msg["Type"] == "SENSORSTATUS":
        updateSensorStatus()
        sendSensorStatus()
        return
    logger.warning("Invalid StatusRequest Type")
    logger.debug(msg)


def getTimestamp():
    return int(time.mktime(datetime.datetime.now().timetuple()))


def getUptime():
    """Get system uptime in seconds"""
    now = datetime.datetime.now()
    current_timestamp = time.mktime(now.timetuple())
    return int(current_timestamp - psutil.boot_time())


def updateSensorStatus():
    """Check for running measurement and update global sensor status"""
    global statusInfo
    global status

    statusInfo = getRunningInfo()
    # Nothing running yet, state is not affected by running process
    # XXX [Markus] What if sensing was started by another process?
    # The dongle allows one application at the same time, so this could
    # lead to problems
    # [Héctor] By design no other process will start sensing. But let's keep
    # these comments in case this design changes in the future.

    if subproc is None:
        return

    # process running (not killed) and status idle
    if subproc.poll() is None and status == "IDLE":
        errorFail("Sensor process running in IDLE state", "ERROR")
        if stopRunningProcess():
            status = "IDLE"
            # FIXME [Markus] ERROR state if this failed?
            # [Héctor] If we implement an error counter at the controller
            # to behave differently it can make sense. If not, being
            # in IDLE or ERROR will have the same outcome (retry from the controller)
        return

    # process not running and status sensing
    if subproc.poll() is not None and status == "SENSING":
        (output, error) = subproc.communicate()
        # Recoverable error by the backend/controller
        errorFail("Sensor process not running in SENSING state:\n" +
                    output + error,
                    "IDLE")


def measureTemperature():
    """Returns temperature value from RPi CPU sensor as string"""
    temp = os.popen("vcgencmd measure_temp").readline().rstrip()
    return (temp.replace("temp=","").replace("'C",""))


def getRunningInfo():
    """Get information about running measurement process (PID, Memory, CPU),
    free disk space and campaign id"""
    runningInfo = {}

    # Treating the current timesynchronization mechanism as a potentially
    # changing value
    timing_metadata_file = "/tmp/time_sync_mode"

    if os.path.isfile(timing_metadata_file):
        with open(timing_metadata_file) as f:
            lines = [line.rstrip('\n') for line in f]
        if len(lines)==1:
            runningInfo["Timing"] = lines[0]

    runningInfo["FreeSpace"] = str(psutil.disk_usage('/'))
    # temperature = str(psutil.sensors_temperatures())
    if subproc is not None:
        if subproc.poll() is None:
            p = psutil.Process(pid=subproc.pid)
            children = p.children()
            # sudo has the running process as only child
            if len(children) > 0:
                p = children[0]
            runningInfo["ProcessID"] = str(p.pid)
            runningInfo["ProcessMemory"] = str(p.memory_info())
            runningInfo["ProcessCPU"] = str(
                p.cpu_percent(interval=None))
    runningInfo["CampaignId"] = campaign
    runningInfo["Temperature"] = measureTemperature()
    return runningInfo


def getPlatformInfo():
    """Get information about installed electrosense Debian packages, kernel,
    distro, libc, CPU info and image release number"""

    apt_pkg.init()
    cache = apt_pkg.Cache()
    sensing_package = ["electrosense-sensor"]
    packages = ["electrosense-sensor", "electrosense-sensor-mqtt", "electrosense-sensor-management"]
    image_metadata_file = "/boot/es_image"
    sensing_version = "Unknown"
    image_version = "Unknown"

    for package in sensing_package:
        if package in cache:
            package_info = cache[package]
            sensing_version = package_info.current_ver.ver_str
            break

    installed_packages = {}
    for package in packages:
        if package in cache:
            package_info = cache[package]
            installed_packages[package] = package_info.current_ver.ver_str

    controlPresent = os.path.isfile("/etc/electrosense/control")

    if os.path.isfile(image_metadata_file):
        with open(image_metadata_file) as f:
            lines = [line.rstrip('\n') for line in f if line.startswith("version")]
        if len(lines)>0:
            image_version = lines[0].replace("version:","")

    platformInfo = {}
    platformInfo["ElectrosensePackage"] = sensing_version
    platformInfo["InstalledPackages"] = str(installed_packages)
    platformInfo["Official"] = str(controlPresent)
    platformInfo["ImageVersion"] = image_version
    platformInfo["Uname"] = str(platform.uname())
    platformInfo["Platform"] = str(platform.platform())
    platformInfo["Distro"] = str(platform.linux_distribution())
    platformInfo["LibC"] = str(platform.libc_ver())
    platformInfo["Cores"] = str(psutil.cpu_count())
    return platformInfo


# Generic AVRO/MQTT message sending function
def sendMessage(msg, topic):
    """Send a message to MQTT topic"""
    (result, mid) = mc.publish(topic, msg, 0)
    logger.debug("Message sent: %s, Result %d, MID: %d",
                 topic, result, mid)


def sendSensorStatus():
    """Publish SensorStatus message to MQTT"""
    logger.debug("Generating SensorStatus")
    msg = {}
    msg["SerialNumber"] = serialNumber
    msg["Timestamp"] = getTimestamp()  # grab this at server side?
    msg["Uptime"] = getUptime()
    msg["Status"] = status
    msg["StatusInfo"] = statusInfo
    data = ma.createAVROMessage("SensorStatus", msg)
    sendMessage(data, "sensor/status/"+clientId)
    logger.info("SensorStatus message sent: %s", status)


def sendSensorInfo():
    """Publish SensorInfo message to MQTT"""
    logger.debug("Generating SensorInfo")
    msg = {}
    msg["SerialNumber"] = serialNumber
    msg["Timestamp"] = getTimestamp()
    msg["SensorType"] = ""
    msg["SoftwareInfo"] = platformInfo
    data = ma.createAVROMessage("SensorInfo", msg)
    sendMessage(data, "sensor/status/"+clientId)
    logger.info("SensorInfo message sent")    


def sendLog(level, text):
    """Publish Log message to MQTT"""
    logger.debug("Generating Log")
    msg = {}
    msg["SerialNumber"] = serialNumber
    msg["Timestamp"] = getTimestamp()
    msg["Level"] = level
    msg["LogMessage"] = text
    data = ma.createAVROMessage("Log", msg)
    sendMessage(data, "sensor/status/"+clientId)
    logger.info("Log message sent")


def mqttConnectedActions():
    """After MQTT connection setup, immediately send sensor status and sensor info"""
    sendSensorStatus()
    sendSensorInfo()


# MQTT event callbacks  

def on_mqtt_connect(client, userdata, flags, rc):
    """The callback for when the client receives a CONNACK response from the server.
    """
    logger.info("Connected to MQTT broker with result code '%s'",
                mqtt.connack_string(rc))
    # Subscribing in on_connect() means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    # client.subscribe("$SYS/#")
    for topic in topics:
        (r, mid) = client.subscribe(topic, 1)
        logger.debug("Subscription message to " + topic +
                     ": " + str(mid) + " " + str(r))
    mqttConnectedActions()

def on_mqtt_disconnect(client, userdata, rc):
    """MQTT event callback when a disconnection is detected"""
    if rc != 0:
        logger.error("Unexpected disconnection. Code %d", rc)


def on_mqtt_message(client, userdata, message):
    """Message handler for incoming MQTT messages"""
    logger.debug("Received message")
    try:
        r = ma.decodeAvroMessage(message.payload)
    except:
        logger.warning("Incorrect AVRO message received")
        return
    if "Type" not in r:
        logger.warning("Incorrect message received")
        return
    mt = r["Type"]
    if mt not in actions:
        logger.warning("Incorrect message type received: %s", mt)
        return
    actions[mt](r["Message"])


def on_mqtt_subscribe(client, obj, mid, granted_qos):
    logger.debug("Subscribed: " + str(mid) +
                 " GrantedQoS: " + str(granted_qos))


def lastWillMessage():
    """Publish final sensor status to MQTT when shutting down"""
    logger.info("Generating LastWillMessage")
    msg = {}
    msg["SerialNumber"] = serialNumber
    msg["Timestamp"] = 0  # shouldn't this be got at server side???
    msg["Uptime"] = 0
    msg["Status"] = "OFF"
    return ma.createAVROMessage("SensorStatus", msg)


""" 
def handleSignal(signum, stack):
    logger.info("Received signal %s", signum)
    if signum==signal.SIGUSR1:
        logger.info("SIGUSR1 received, WebRTC taking control") """


def startWebRTCMode():
    params = {"command": "es_sensor",
                "minFreq": "127000000",
                "maxFreq": "127000000",
                "measurementType": "webrtc",
                "CampaignId": "WebRTC"}
    msg = {"Uuid": "local_action", 
            "Command": "START", 
            "SensingParams": params,
            "Timestamp": getTimestamp()}
    processMeasurementCommand(msg)


def stopWebRTCMode():
    msg = {"Uuid": "local_action", 
           "Command": "START",
           "SensingParams": {},
           "Timestamp": getTimestamp()}
    processMeasurementCommand(msg)


# INITS

# BIG TODO This status info block should be moved
# from dirty global variables to an object

# apply serial number from config if present, otherwise take MAC address
if "serialnumber" in envconfig.keys():
    sensorMac = envconfig["serialnumber"]
    logger.info("Using sensorMac from config file serialnumber: "+str(sensorMac))
else:
    try:
        ifaddr = netifaces.ifaddresses(envconfig["interface"])
        sensorMac = str(ifaddr[netifaces.AF_LINK][0]['addr'])
    except ValueError:
        logger.error("Could not read MAC address for interface "
                     + envconfig["interface"])
        sys.exit(-1)
# If userif is defined it overrides the detected value
translation = str.maketrans('', '', ":.- ")
serialNumber = int(sensorMac.translate(translation), 16)
platformInfo = getPlatformInfo()
# ClientID will be the detected clientId, overrided by
# the config userid if present:
if "userid" in mqttconfig.keys():
    clientId = mqttconfig["userid"]
else:
    clientId = str(serialNumber)
groups = set()
status = "OFF"
statusInfo = {}
campaign = "Unknown"
subproc = None
topics = ["control/sensor/all", "control/sensor/id/"+clientId]

# The callback dict for when a message is received
actions = {"MeasurementCommand": processMeasurementCommand,
           "GroupConfiguration": processGroupIDs,
           "StatusRequest": processStatusRequest}


# AVRO init
ma.init(avroconfig)

logger.info("MAC: " + sensorMac)
logger.info("CID: " + clientId)
mc = mqtt.Client(client_id=clientId,
                 clean_session=True,
                 userdata=None)

mc.username_pw_set(username=mqttconfig["user"],
                   password=mqttconfig["password"])
mc.tls_set(
    ca_certs=mqttconfig["ca"],
    certfile=mqttconfig["certificate"],
    keyfile=mqttconfig["key"],
    tls_version=2)

mc.on_connect = on_mqtt_connect
mc.on_message = on_mqtt_message
mc.on_subscribe = on_mqtt_subscribe
mc.on_disconnect = on_mqtt_disconnect

# Message to be sent when an uncontrolled disconnection occurs
mc.will_set("sensor/status/"+clientId, payload=lastWillMessage(), qos=1)

mc.max_inflight_messages_set(int(mqttconfig["inflight"]))

try:
    mc.connect(mqttconfig["server"], int(mqttconfig["port"]), 60)
except Exception as err:
    # Paho returns socket.error exception, why do they
    # force to import another module just for an Exception???
    print("Could not connect to MQTT broker. Check MQTT configuration.")
    print(err)
    sys.exit(1)

# Blocking call that processes network traffic, dispatches callbacks and
# handles reconnecting.
# Other loop*() functions are available that give a threaded interface and a
# manual interface.
random.seed()
status = "IDLE"

# use loop_start instead of loop_forever, so it does not block
# loop_start manages reconnections as well
mc.loop_start()

# Set up WebRTC controlling socket
try:
    socketFileName = "/tmp/MQTT_WEBRTC"
    webrtcSocket=socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
    if os.path.exists(socketFileName):
        os.remove(socketFileName)
    webrtcSocket.bind(socketFileName)
    webrtcSocket.listen(1)
    webrtcSocket.setblocking(False)
    #(webrtcConn, webrtcAddr) = webrtcSocket.accept()
except OSError as errormsg:
    webrtcSocket.close()
    logger.error("Could not open WebRTC socket: %s", str(errormsg))


# Implement secondary loop to process sensing status and output
# so information and reaction time is minimised
while True:
    time.sleep(1)
    # Thanks, Python ordered lazy evaluation!
    # In order: are we connected? is there a sensing process? and is it running?
    # TODO we shouldn't empty the buffer if we do not have a connection, or face handling
    # a local log buffering ourselves (timestamped)
    if mc.socket() is not None and subproc is not None and subproc.poll() is None:
            output = '\n'.join(subproc.stdout.readlines())
            error = '\n'.join(subproc.stderr.readlines())
            if len(output)>0:
                sendLog("DEBUG", output)
                logger.debug(output)
            if len(error)>0:
                sendLog("ERROR", error)
                logger.error(error)
    if webrtcSocket is not None and webrtcSocket.fileno() != -1:
        try:
            if "webrtcConn" not in locals() or webrtcConn is None or webrtcConn.fileno() == -1:
                (webrtcConn, webrtcAddr) = webrtcSocket.accept()
            msg = webrtcConn.recv(1024)
            if msg==b"CONNECTED":
                logger.info("WebRTC socket received a connection command.")
                startWebRTCMode()
            elif msg==b"DISCONNECTED":
                logger.info("WebRTC socket received a disconnection command.")
                stopWebRTCMode()
            elif msg!=b'':
                logger.error("WebRTC socket received an unknown command #%s#.", str(msg))
            else:
                webrtcConn.close()
        except OSError:
            pass
